#!/usr/bin/env python3
"""
This file is part of decopy testdata. It exists for testing purposes only.

Copyright: 2016, Jane Doe <janedoe@example.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.
"""



# This file is part of decopy testdata. It exists for testing purposes only.

# Copyright: 2016, Jane Doe <janedoe@example.com>

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.



#!/usr/bin/env python3
# -*- coding: utf-8 -*- vim60:fdm=marker
#
# Copyright: 2016, Maximiliano Curia <maxy@debian.org>
#
# License: ISC
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
#  AND FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
#  LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
#  OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
#  PERFORMANCE OF THIS SOFTWARE.

import logging
import sys

from contextlib import contextmanager

from .datatypes import License


@contextmanager
def open_output_file(filename):
    if filename:
        with open(filename, 'wt', encoding='utf-8') as f:
            yield f
    else:
        yield sys.stdout


def _generate_header(copyright_, f, options):
    current_format = (
        'https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/'
    )
    if options.mode == 'partial':
        paragraph = (
            'Format: {}\n'
            'Source: TODO\n'
            'Comment: *** only: {} ***'.format(
                current_format, ', '.join(options.files)))
    elif copyright_:
        header = copyright_.header
        header.format = current_format
        if not header.source:
            header.source = 'TODO'
        paragraph = header.dump().rstrip('\n')
    else:
        paragraph = 'Format: {}\nSource: TODO'.format(current_format)
    print(paragraph, file=f)
    if options.output:
        logging.debug('Generated header:\n%s', paragraph)


def generate_output(groups, filetree, copyright_, options):

    # Track the licenses that are in use
    licenses = set()

    with open_output_file(options.output) as f:
        # Print header
        _generate_header(copyright_, f, options)

        # Avoid printing 'Files: *' when working in partial mode
        if options.mode == 'partial':
            for item in options.files:
                if not filetree[item].parent:
                    continue
                filetree[item].parent.tally()

        # Print files paragraphs
        for _, group in sorted(
                groups.items(), key=lambda i: i[1].sort_key(options)):

            if not group.copyright_block_valid():
                continue

            licenses.update(group.licenses.keys())

            paragraph = group.copyright_block(options.glob)

            print(file=f)
            print(paragraph, file=f)

            if options.output:
                logging.debug('Generated group:\n%s', paragraph)

        # Print license paragraphs
        for key in sorted(licenses):
            license_ = License.get(key)
            paragraph = str(license_)

            print(file=f)
            print(paragraph, file=f)

            if options.output:
                logging.debug('Generated license block:\n%s', paragraph)



#!/usr/bin/env python3
# -*- coding: utf-8 -*- vim60:fdm=marker
#
# Copyright: 2016, Maximiliano Curia <maxy@debian.org>
#
# License: ISC
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
#  AND FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
#  LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
#  OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
#  PERFORMANCE OF THIS SOFTWARE.

''' Miscelaneous data types '''

try:
    import regex as re
except ImportError:
    import re

from collections import namedtuple

ReSub = namedtuple('ReSub', ('pattern', 'repl'))
ReFlags = namedtuple('ReFlags', ('re', 'flags'))
ReFlags.__new__.__defaults__ = (None, re.IGNORECASE)
ReLicense = namedtuple('ReLicense', ('re', 'license', 'get_detail', 're_flags'))
ReLicense.__new__.__defaults__ = (None, None, None, re.IGNORECASE)

UNKNOWN = 'Unknown'
UNKNOWN_COPYRIGHTED = 'UnknownCopyrighted'


class License(object):

    __slots__ = ('name', 'stored')

    licenses = {}

    def __init__(self, name):
        self.name = name
        self.stored = None

    def __str__(self):
        if self.stored:
            return self.stored.dump().rstrip('\n')
        return ('License: {name}\n'
                'Comment: Add the corresponding license text here'.format(
                    name=self.name))

    @staticmethod
    def get(name):
        if name not in License.licenses:
            License.licenses[name] = License(name)

        return License.licenses[name]


class YearRange(object):

    __slots__ = ('low', 'high')

    def __init__(self, low=None, high=None):
        if low is None:
            low_value = 0
        else:
            low_value = int(low)

        if high is None:
            high_value = low_value
        else:
            high_value = int(high)

        if low_value > high_value:
            low_value, high_value = high_value, low_value
        self.low = low_value
        self.high = high_value

    def __in__(self, year):
        value = int(year)
        return self.low <= value <= self.high

    def add(self, year):
        value = int(year if year is not None else 0)
        if not value:
            return self
        if not self.low or value < self.low:
            self.low = value
        if not self.high or self.high < value:
            self.high = value
        return self

    def newer(self, other):
        if self.high and other.high:
            return other.high > self.high
        return not self.high and other.high

    def merge(self, other):
        self.add(other.low)
        self.add(other.high)
        return self

    def __bool__(self):
        return bool(self.low)

    def __str__(self):
        if not self.low:
            return ''
        if self.low == self.high:
            return str(self.low)
        return str(self.low) + '-' + str(self.high)


class CopyrightHolder(object):

    __slots__ = ('name', 'email', 'years')

    def __init__(self, name, email, years):
        self.name = name
        self.email = email
        self.years = years if years is not None else YearRange()

    def merge(self, other):
        if other.name and self.years.newer(other.years):
            self.name = other.name
        self.years.merge(other.years)
        return self

    @property
    def person(self):
        result = self.name
        if self.name and self.email:
            result += ' '
        if self.email:
            result += '<{}>'.format(self.email)
        return result

    def __str__(self):
        result = str(self.years)
        result += ', ' if result else ''
        result += self.person
        return result

    def __repr__(self):
        return str(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*- vim60:fdm=marker
#
# Copyright: 2016, Maximiliano Curia <maxy@debian.org>
#
# License: ISC
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
#  AND FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
#  LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
#  OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
#  PERFORMANCE OF THIS SOFTWARE.

''' DEP-5 copyright handling '''

import fnmatch
import logging
import os

try:
    import regex as re
except ImportError:
    import re

import debian.copyright as dc

# Local modules
from .tree import DirInfo, FileGroup, CopyrightGroup
from .matchers import get_copyright_holders
from .datatypes import License, UNKNOWN, UNKNOWN_COPYRIGHTED


class Copyright(dc.Copyright):
    '''This class extends the debian.copyright.Copyright class, adding some
    mangling of paragraphs and the capability of interacting with the filetree
    that is the basis of how decopy works.'''

    groups = None

    @property
    def paragraphs(self):
        return self.__paragraphs

    @staticmethod
    def build(filetree, options):
        '''Creates an instance of Copyright and then populates the filetree.'''

        filename = options.copyright_file
        if not os.path.isabs(filename):
            filename = os.path.join(filetree.root, filename)

        if not os.path.exists(filename):
            return Copyright()

        try:
            with open(filename) as c:
                copyright_ = Copyright(c, strict=False)
        except (dc.NotMachineReadableError, ValueError):
            return Copyright()

        return copyright_

    def process(self, filetree):
        '''Match files in the filetree to copyright information.'''
        files_paragraphs = []

        # Add the license paragraphs to the License instances
        for paragraph in self.all_license_paragraphs():
            license_ = License.get(paragraph.license.synopsis)
            license_.stored = paragraph

        # Scan the files paragraphs in reverse order, that way the files should
        # match only once.
        paragraphs = list(self.all_files_paragraphs())
        for paragraph in reversed(paragraphs):
            # apply fnmatch to each file_pattern
            matched_patterns, files = \
                self._get_matched_patterns(paragraph, filetree)

            # Nothing in the paragraph matched any files. The paragraph will not
            # be included in files_paragraphs
            # TODO: for diff mode
            if not matched_patterns:
                logging.info('No matching pattern in:\n%s', paragraph.dump())
                continue

            # Some patterns (but not all) failed to match.
            # Those specific patterns get removed from the paragraph.
            # TODO: for diff mode
            if len(matched_patterns) != len(paragraph.files):
                paragraph.files = matched_patterns

            files_paragraphs.append((paragraph, files))

        self._process_groups(filetree, files_paragraphs)

    @staticmethod
    def _get_matched_patterns(paragraph, filetree):
        '''Returns the files that match the patterns for a paragraph.'''
        matched_patterns = []
        files = set()

        for file_pattern in paragraph.files:
            found = False
            for filename in fnmatch.filter(filetree.names, file_pattern):
                fileinfo = filetree[filename]

                # If the file already has a matching pattern, it was already
                # matched by previous paragraphs.
                if fileinfo.matching_pattern:
                    continue

                # Set the matched pattern
                fileinfo.matching_pattern = file_pattern
                if not isinstance(fileinfo, DirInfo):
                    files.add(filename)

                if not found:
                    matched_patterns.append(file_pattern)
                    found = True

            if not found:
                # TODO: this might be interesting to print in the diff mode
                logging.info('No match found for %s', file_pattern)

        return matched_patterns, files

    def _process_groups(self, filetree, files_paragraphs):

        self.groups = []

        for position, (paragraph, files) in enumerate(files_paragraphs):

            group = Group(position=position)
            group.stored = paragraph

            # Get license keys from the synopsis
            license_keys = re.split(' or ', paragraph.license.synopsis,
                                    flags=re.I)
            for key in license_keys:
                group.licenses[key] = License.get(key)
            # Does this file paragraph includes the license text?
            # Can we reuse it?
            if len(license_keys) == 1 \
                    and paragraph.license.text:
                license_ = License.get(license_keys[0])
                if not license_.stored:
                    license_.stored = \
                        dc.LicenseParagraph.create(paragraph.license)

            holders = []
            if paragraph.copyright:
                for value in paragraph.copyright.split('\n'):
                    value = value.lstrip()
                    new_holders, _ = get_copyright_holders(value)
                    holders.extend(new_holders)
            group.add_copyrights(holders)

            for filename in files:
                fileinfo = filetree[filename]
                fileinfo.stored_group = group
                group.add_file(fileinfo)

            if paragraph.comment:
                group.comments.append(paragraph.comment)

            self.groups.append(group)

    def remove_misplaced_files(self, options):
        '''Remove files from groups where they no longer belong.

        This method should be called after files have been processed and their
        licensing and copyright information has been obtained. Files that do not
        match their placement in the copyright file get removed from those
        groups. No files are added to existing or new groups.
        '''
        for group in self.groups:
            group_key = group.get_key(options)
            group_key_no_copyright = group.get_key(options, ignore_copyright=True)
            logging.debug("Keys: %s %s", group_key, group_key_no_copyright)

            # Make an explicit copy to be able to remove items from the list.
            for fileinfo in list(group.files):
                # In partial mode, only included files are taken into account
                if options.mode == 'partial' and not fileinfo.included:
                    group.files.del_file(fileinfo)
                    continue

                # Check if the current key for the file matches the group key.
                file_key = fileinfo.get_group_key(options)
                logging.debug("File: %s %s", fileinfo, file_key)
                remove = file_key != group_key

                # options.group_by copyright would split the parts when the
                # copyright holder is not found
                if remove and options.group_by == 'copyright':
                    remove = file_key != group_key_no_copyright

                # Either delete from the group or set the group field in the
                # fileinfo, to show that the file has matched its group.
                if remove:
                    # TODO: for diff mode
                    group.files.del_file(fileinfo)
                else:
                    # Update parsed copyrights
                    group.copyrights.merge(fileinfo.copyrights)
                    fileinfo.group = group

    def get_group_dict(self, options):
        groups = {}
        for group in self.groups:
            key = group.get_key(options)
            if key in groups:
                for fileinfo in group.files:
                    groups[key].add_file(fileinfo)
            else:
                groups[key] = group
        return groups


class Group(object):

    def __init__(self, key=None, position=0):
        self.key = key
        self.files = FileGroup()
        self.copyrights = CopyrightGroup()
        self.licenses = {}
        self.license_filenames = set()
        self.comments = []
        self.stored = None
        self.position = position

    @property
    def license(self):
        name = ' or '.join(sorted(self.licenses))
        if not name:
            if self.copyrights:
                return UNKNOWN_COPYRIGHTED
            return UNKNOWN
        return name

    def __str__(self):
        return str(self.key) if self.key else self.license

    def __repr__(self):
        return 'Group({})'.format(str(self))

    def add_file(self, fileinfo):
        self.files.add_file(fileinfo)
        self.copyrights.merge(fileinfo.copyrights)
        for name in fileinfo.get_licenses():
            self.licenses[name] = License.get(name)
        self.license_filenames.update(fileinfo.get_license_filenames())

    def add_copyrights(self, copyrights):
        self.copyrights.extend(copyrights)

    def get_inherited_comment(self):
        if not self.license_filenames:
            return
        # TODO: Is this enough?
        if self.stored:
            return
        filenames = sorted(self.license_filenames)
        return 'No explicit license found, using license(s) from:\n {}'.format(
            '\n '.join(filenames))

    def get_comments(self):
        comments = []
        inherited_comment = self.get_inherited_comment()
        if inherited_comment:
            comments.append(inherited_comment)
        comments.extend(self.comments)
        return '\n .\n '.join(comments)

    def get_license_key(self):
        return self.license

    def get_copyright_key(self):
        return self.copyrights.key()

    def get_path_key(self, options):
        return self.files.key(options)

    def get_key(self, options, ignore_copyright=False):
        if not self.key:
            license_key = self.get_license_key()

            if options.group_by == 'copyright':
                copyright_key = self.get_copyright_key()
            else:
                copyright_key = None

            path_key = self.get_path_key(options)

            self.key = (license_key, copyright_key, path_key)
        if ignore_copyright:
            return self.key[0], tuple(), self.key[2]
        return self.key

    def copyright_block_valid(self):
        if not self.files:
            return False
        if len(self.files) > 1:
            return True
        return bool(list(f.fullname for f in self.files)[0])

    def copyright_block(self, glob_):
        if not self.copyright_block_valid():
            return ''

        if glob_:
            files = '\n       '.join(self.files.get_patterns())
        else:
            # When called with --no-glob, only list the file names
            files = '\n       '.join(self.files.sorted_members())

        block = ['Files: {files}'.format(files=files)]
        if self.copyrights:
            holders = '\n           '.join(self.copyrights.sorted_members())
            block.append('Copyright: {holders}'.format(holders=holders))
        block.append('License: {license}'.format(license=self.license))
        comments = self.get_comments()
        if comments:
            block.append('Comment: {}'.format(comments))
        # And finally add all the "unknown" fields
        if self.stored:
            for field in self.stored:
                if field in {'Files', 'Copyright', 'License', 'Comment'}:
                    continue
                block.append('{}: {}'.format(field, self.stored[field]))
        return '\n'.join(block)

    def sort_key(self, options):
        debian_group = 0
        if options.split_debian:
            path_key = self.get_path_key(options)
            if path_key == 'debian':
                debian_group = 2

        # sort unknown licenses below others
        if self.license == UNKNOWN or self.license == UNKNOWN_COPYRIGHTED:
            debian_group = 1

        return debian_group, -1 * self.position, -1 * len(self.files), -1 * len(self.copyrights), self.get_key(options, ignore_copyright=True)



#!/usr/bin/env python3
# encoding: utf-8

# Unit tests for decopy
# Copyright (C) 2016 Margarita Manterola <marga@debian.org>
#
# License: ISC
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
#  AND FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
#  LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
#  OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
#  PERFORMANCE OF THIS SOFTWARE.

import unittest

from decopy import cmdoptions
from decopy import tree
from decopy import dep5
from decopy.datatypes import CopyrightHolder, YearRange


class TestFileInfo(unittest.TestCase):

    def setUp(self):
        # Create one root with one dir with one file
        self.r = tree.RootInfo(root='root')
        self.d = tree.DirInfo(name='dir', parent=self.r)
        self.f = tree.FileInfo(name='file', parent=self.d)

    def testFullname(self):
        self.assertEqual(self.r.fullname, '')
        self.assertEqual(self.d.fullname, 'dir')
        self.assertEqual(self.f.fullname, 'dir/file')

    def testGetLicenses(self):
        # Set a license for the directory
        self.d.add_licenses({'GPL'}, 'COPYING')
        # The license for the file should be the one from the directory
        self.assertEqual(self.f.get_licenses(), {'GPL'})

        # Set a different license for the file
        self.f.add_licenses({'BSD'})
        # Now the license is the one for the file
        self.assertEqual(self.f.get_licenses(), {'BSD'})

        # Add a second license to the file
        self.f.add_licenses({'PHP'})
        # Now it should have both licenses
        self.assertEqual(self.f.get_licenses(), {'BSD', 'PHP'})

    def testSplittingPath(self):
        # Set a license for the directory
        self.d.add_licenses({'GPL'}, 'COPYING')
        # Create a child directory with a different license
        self.o = tree.DirInfo(name='other', parent=self.d)
        self.o.add_licenses({'BSD'}, 'COPYING')

        # First test with splitting on license
        options = cmdoptions.process_options(['--no-progress', '--split-on-license'])
        self.assertEqual(self.d.get_splitting_path(options), 'dir')
        self.assertEqual(self.o.get_splitting_path(options), 'dir/other')

        # Then test without splitting on license
        options = cmdoptions.process_options(['--no-progress', '--no-split-on-license'])
        self.assertEqual(self.d.get_splitting_path(options), '')
        self.assertEqual(self.o.get_splitting_path(options), '')

        # Create a debian directory, with another license
        self.deb = tree.DirInfo(name='debian', parent=self.r)
        self.deb.add_licenses({'LGPL'}, 'debian/copyright')
        options = cmdoptions.process_options(
            ['--no-progress', '--split-debian', '--no-split-on-license'])
        self.assertEqual(self.deb.get_splitting_path(options), 'debian')

        options = cmdoptions.process_options(
            ['--no-progress', '--split-debian', '--split-on-license'])
        self.assertEqual(self.deb.get_splitting_path(options), 'debian')

        options = cmdoptions.process_options(
            ['--no-progress', '--no-split-debian', '--split-on-license'])
        self.assertEqual(self.deb.get_splitting_path(options), 'debian')

        options = cmdoptions.process_options(
            ['--no-progress', '--no-split-debian', '--no-split-on-license'])
        self.assertEqual(self.deb.get_splitting_path(options), '')

    def testTally(self):
        self.assertEqual(self.f.get_first_untallied(), self.r)
        self.d.tally()
        self.assertEqual(self.f.get_first_untallied(), self.f)

    def testIncluded(self):
        self.assertEqual(self.f.included, False)
        # Remove the cache
        self.f.included = None
        self.d.included = True
        self.assertEqual(self.f.included, True)


class TestRootInfo(unittest.TestCase):

    def _check_expected_files(self, filetree, expected_files):
        '''Check that the parsed filetree has the expected files.'''
        self.longMessage = False
        for filename in expected_files:
            with self.subTest(msg='Checking "{}"'.format(filename)):
                try:
                    name = filetree[filename].fullname
                except IndexError:
                    name = ''
                self.assertEqual(name, filename,
                                 msg='missing file: "{}"'.format(filename))

        for fileinfo in filetree:
            with self.subTest(msg='Checking "{}"'.format(fileinfo.fullname)):
                self.assertTrue(
                    fileinfo.fullname in expected_files,
                    msg='unexpected file: "{}"'.format(fileinfo.fullname))
        self.longMessage = True

    def testSimpleTree(self):
        '''Parses a tree with licensed files (testdata/one).'''

        options = cmdoptions.process_options(
            ['--no-progress', '--root', 'tests/testdata/one'])
        expected_files = {
            '': 'Inherited(GPL-2+)',
            'COPYING': 'GPL-2+',
            'art.txt': 'Artistic-1',
            'bsd.txt': 'BSD-3-clause',
            'empty.txt': 'Inherited(GPL-2+)',
            'no.txt': 'Inherited(GPL-2+)',
            'paren.txt': 'GPL-3+',
        }

        # First build the tree and check that it has the files and nothing else
        filetree = tree.RootInfo.build(options)
        self._check_expected_files(filetree, expected_files)

        # Now process the licenses
        filetree.process(options)

        # Check that the licenses match our expectations
        for filename, license_ in expected_files.items():
            with self.subTest(msg='Checking "{}"'.format(filename)):
                self.assertEqual(
                    filetree[filename].get_license_key(), license_,
                    msg='license mismatch for: "{}"'.format(filename))

    def testCopyrightedTree(self):
        '''Parses a tree with licensed & copyrighted files (testdata/two).'''

        options = cmdoptions.process_options(
            ['--no-progress', '--root', 'tests/testdata/two'])
        expected_files = {
            '': ['Inherited(GPL-2+)', ()],
            'COPYING': ['GPL-2+', ()],
            'art.txt': ['Artistic-1', ('Jane Doe',)],
            'bsd.txt': ['BSD-3-clause', ('John Doe',)],
            'empty.txt': ['Inherited(GPL-2+)', ()],
            'no.txt': ['Inherited(GPL-2+)', ('Juan Pérez', 'Juana López')],
            'docbook.txt': ['Inherited(GPL-2+)', ()],
        }

        # First build the tree and check that it has the files and nothing else
        filetree = tree.RootInfo.build(options)
        self._check_expected_files(filetree, expected_files)

        # Now process the licenses
        filetree.process(options)

        # Check that the licenses and copyrights match our expectations
        for filename, values in expected_files.items():
            license_, copyright_ = values
            with self.subTest(msg='Checking "{}"'.format(filename)):
                self.assertEqual(
                    filetree[filename].get_license_key(), license_,
                    msg='license mismatch for: "{}"'.format(filename))

                self.assertEqual(
                    filetree[filename].get_copyright_key(),
                    tuple(sorted(copyright_)),
                    msg='copyright mismatch for: "{}"'.format(filename))

    def testDebianizedTree(self):
        '''Parses a tree that contains a debian/copyright (testdata/three).'''

        options = cmdoptions.process_options(
            ['--no-progress', '--root', 'tests/testdata/three'])
        expected_files = {
            '': ['Inherited(GPL-2+)', ()],
            'COPYING': ['GPL-2+', ()],
            'src': ['Inherited(GPL-2+)', ()],
            'src/do-something.m': ['MIT/X11', ('clara@example.com',)],
            'src/do-something.py': ['MIT/X11', ('clara@example.com',)],
            'src/main.py': ['GPL-2+', ('janedoe@example.com',)],
            'src/module.py': ['GPL-2+', ('janedoe@example.com',)],
            'acinclude.m4': ['GPL-2+', ('Free Software Foundation, Inc',)],
            'data': ['Inherited(GPL-2+)', ()],
            'data/file.txt': ['MIT/X11', ('clara@example.com',)],
            'debian': ['Inherited(GPL-2+)', ()],
            'debian/rules': ['LGPL-2+', ()],
            'debian/compat': ['LGPL-2+', ()],
            'debian/copyright': ['LGPL-2+', ()],
        }

        # First build the tree and check that it has the files and nothing else
        filetree = tree.RootInfo.build(options)
        self._check_expected_files(filetree, expected_files)

        # Instantiate and process the copyright
        copyright_ = dep5.Copyright.build(filetree, options)
        copyright_.process(filetree)

        # Now process the licenses in the filetree
        filetree.process(options)

        # Check that the licenses and copyrights match our expectations
        for filename, values in expected_files.items():
            license_, copyright_ = values
            with self.subTest(msg='Checking "{}"'.format(filename)):
                self.assertEqual(
                    filetree[filename].get_license_key(), license_,
                    msg='license mismatch for: "{}"'.format(filename))

                self.assertEqual(
                    filetree[filename].get_copyright_key(),
                    tuple(sorted(copyright_)),
                    msg='copyright mismatch for: "{}"'.format(filename))

    def testTreeExclude(self):
        '''Parse a tree with an exclusion expresion (testdata/three).'''

        options = cmdoptions.process_options(
            ['--no-progress', '--root', 'tests/testdata/three', '-X', 'src.*'])
        expected_files = {
            '': ['Inherited(GPL-2+)', ()],
            'COPYING': ['GPL-2+', ()],
            'src': ['Inherited(GPL-2+)', ()],
            'src/do-something.m': ['Inherited(GPL-2+)', ()],
            'src/do-something.py': ['Inherited(GPL-2+)', ()],
            'src/main.py': ['Inherited(GPL-2+)', ()],
            'src/module.py': ['Inherited(GPL-2+)', ()],
            'acinclude.m4': ['GPL-2+', ('Free Software Foundation, Inc',)],
            'data': ['Inherited(GPL-2+)', ()],
            'data/file.txt': ['MIT/X11', ('clara@example.com',)],
            'debian': ['Inherited(GPL-2+)', ()],
            'debian/rules': ['Inherited(GPL-2+)', ()],
            'debian/compat': ['Inherited(GPL-2+)', ()],
            'debian/copyright': ['Inherited(GPL-2+)', ()],
        }

        # First build the tree and check that it has the files and nothing else
        filetree = tree.RootInfo.build(options)
        self._check_expected_files(filetree, expected_files)

        # Don't process the debian/copyright, so that we can test just the
        # parsed tree.

        # Now process the licenses in the filetree. Excluded files won't get
        # processed, so they get inherited values.
        filetree.process(options)

        # Check that the licenses and copyrights match our expectations
        for filename, values in expected_files.items():
            license_, copyright_ = values
            with self.subTest(msg='Checking "{}"'.format(filename)):
                self.assertEqual(
                    filetree[filename].get_license_key(), license_,
                    msg='license mismatch for: "{}"'.format(filename))

                self.assertEqual(
                    filetree[filename].get_copyright_key(),
                    tuple(sorted(copyright_)),
                    msg='copyright mismatch for: "{}"'.format(filename))


class TestFileGroup(unittest.TestCase):

    def setUp(self):
        self.group = tree.FileGroup()

    def testGetPatternsEmptyGroup(self):
        self.assertEqual(self.group.get_patterns(), [])

    def testGetPatternsOneFile(self):
        fileinfo = tree.FileInfo(None, 'foo')
        self.group.add_file(fileinfo)
        self.assertEqual(self.group.get_patterns(), ['foo'])

    def testGetPatternsOneDirOneFile(self):
        dirinfo = tree.DirInfo(None, 'dir')
        fileinfo = tree.FileInfo(dirinfo, 'foo')
        self.group.add_file(fileinfo)
        self.assertEqual(self.group.get_patterns(), ['dir/*'])

    def testGetPatternsSeparateGroups(self):
        dirinfo = tree.DirInfo(None, 'dir')
        foo = tree.FileInfo(dirinfo, 'foo')
        bar = tree.FileInfo(dirinfo, 'bar')
        self.group.add_file(foo)
        other = tree.FileGroup()
        other.add_file(bar)
        self.assertEqual(self.group.get_patterns(), ['dir/*'])
        self.assertEqual(other.get_patterns(), ['dir/bar'])

    def testGetPatternsRoot(self):
        dirinfo = tree.RootInfo('root')
        fileinfo = tree.FileInfo(dirinfo, 'foo')
        self.group.add_file(fileinfo)
        self.assertEqual(self.group.get_patterns(), ['*'])

    def testGetPatternsSeparateGroupsMoreFiles(self):
        dirinfo = tree.DirInfo(None, 'dir')
        foo = tree.FileInfo(dirinfo, 'foo')
        bar = tree.FileInfo(dirinfo, 'bar')
        baz = tree.FileInfo(dirinfo, 'baz')
        self.group.add_file(foo)
        other = tree.FileGroup()
        other.add_file(bar)
        other.add_file(baz)
        self.assertEqual(self.group.get_patterns(), ['dir/*'])
        self.assertEqual(other.get_patterns(), ['dir/bar', 'dir/baz'])

    def testCommonPathEmpty(self):
        self.assertEqual(self.group.commonpath(), '')

    def testCommonPathOneFile(self):
        fileinfo = tree.FileInfo(None, 'foo')
        self.group.add_file(fileinfo)
        self.assertEqual(self.group.commonpath(), 'foo')

    def testCommonPathOneDirTwoFiles(self):
        dirinfo = tree.DirInfo(None, 'dir')
        foo = tree.FileInfo(dirinfo, 'foo')
        bar = tree.FileInfo(dirinfo, 'bar')
        self.group.add_file(foo)
        self.group.add_file(bar)
        self.assertEqual(self.group.commonpath(), 'dir')


class TestCopyrightGroup(unittest.TestCase):

    def setUp(self):
        self.group = tree.CopyrightGroup()

    def testAddTwoPeople(self):
        self.group.add(
            CopyrightHolder('Jane Doe', 'jane@example.com', YearRange(2015, 2015)))
        self.group.add(
            CopyrightHolder('John Doe', 'john@example.com', YearRange(2016, 2016)))

        self.assertEqual(list(self.group.sorted_members()),
                         ['2015, Jane Doe <jane@example.com>',
                          '2016, John Doe <john@example.com>'])

    def testMergeNames(self):
        self.group.add(
            CopyrightHolder('Jane', 'jane@example.com', YearRange()))
        self.group.add(
            CopyrightHolder('Jane Doe', 'jane@example.com', YearRange()))

        self.assertEqual(list(self.group.sorted_members()),
                         ['Jane <jane@example.com>'])

    def testMergeEmails(self):
        self.group.add(
            CopyrightHolder('Jane Doe', 'jane@example.com', YearRange()))
        self.group.add(
            CopyrightHolder('Jane Doe', 'jane@mailinator.com', YearRange()))

        self.assertEqual(list(self.group.sorted_members()),
                         ['Jane Doe <jane@example.com>',
                          'Jane Doe <jane@mailinator.com>'])

    def testMergeYears(self):
        self.group.add(
            CopyrightHolder('Jane Doe', 'jane@example.com', YearRange(2010, 2011)))
        self.group.add(
            CopyrightHolder('Jane Doe', 'jane@example.com', YearRange(2015, 2016)))

        self.assertEqual(list(self.group.sorted_members()),
                         ['2010-2016, Jane Doe <jane@example.com>'])

    def testMergeNamesWithYears(self):
        self.group.add(
            CopyrightHolder('Jane Doe', 'jane@example.com', YearRange(1999, 1999)))
        self.group.add(
            CopyrightHolder('Janine Doe', 'jane@example.com',
                            YearRange(2016, 2016)))

        self.assertEqual(list(self.group.sorted_members()),
                         ['1999-2016, Janine Doe <jane@example.com>'])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*- vim60:fdm=marker
#
# Copyright: 2016, Maximiliano Curia <maxy@debian.org>
#
# License: ISC
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
#  AND FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
#  LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
#  OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
#  PERFORMANCE OF THIS SOFTWARE.

''' Files and Directories information '''

import logging
import os

try:
    import regex as re
except ImportError:
    import re

try:
    from tqdm import tqdm
except ImportError:
    def tqdm(*a, **kw):
        if a:
            return a[0]
        return kw.get('iterable', None)

from concurrent import futures

from .datatypes import UNKNOWN, UNKNOWN_COPYRIGHTED
from .parsers import parse_file


# Compatibility function - Python 3.4 doesn't have os.path.commonpath
if not hasattr(os.path, "commonpath"):
    def commonpath(paths):
        prefix = os.path.commonprefix(paths)
        if not os.path.exists(prefix):
            prefix = prefix.rsplit('/', 1)[0]
        return prefix
    os.path.commonpath = commonpath


class FileInfo(object):

    def __init__(self, parent, name):
        self.parent = parent
        self.name = name
        self.licenses = set()
        self.license_filenames = set()
        # Keep track of files that are assigned to a group
        self.group = None
        # Value retrieved from the debian/copyright file
        self.stored_group = None
        # The pattern in the debian/copyright that matches this file
        self.matching_pattern = None
        self.copyrights = CopyrightGroup()
        # part of the wildcard logic, was this element a partially
        # printed?
        self.tallied = False
        # Store whether file was included in the files/dirs to process
        self._included = None

    def __str__(self):
        return self.fullname

    @property
    def fullname(self):
        if self.parent:
            return os.path.join(self.parent.fullname, self.name)
        return self.name

    @property
    def parsed_license(self):
        return ' or '.join(sorted(self.licenses))

    def add_copyrights(self, copyrights):
        self.copyrights.extend(copyrights)

    def add_licenses(self, licenses, filenames=None):
        '''Add one or more detected licenses for this file.

           Args:
               licenses: a set of license keys.
               filenames: for licenses stored in separate files, the filenames
                   hold the license information.
        '''
        self.licenses.update(licenses)
        if filenames:
            self.license_filenames.update(filenames)

    def get_licenses(self):
        if self.licenses or not self.parent:
            return self.licenses
        return self.parent.get_licenses()

    def get_license_filenames(self):
        if self.licenses or not self.parent:
            return self.license_filenames
        return self.parent.get_license_filenames()

    def get_license_key(self):
        '''Identifier corresponding to the license to be assigned to the file.
           The logic to choose the license is quite complex. This is the order
           for choosing:
            1 - An explicit entry in debian/copyright (i.e. not wildcarded)
            2 - The license detected when parsing the file
            3 - The license stored in debian/copyright (wildcarded)
            4 - The license for the containing directory.

           If none of those is found, then the file is marked as UNKNOWN.
        '''
        if self.stored_group and self.fullname == self.matching_pattern:
            return self.stored_group.get_license_key()
        if self.parsed_license:
            return self.parsed_license
        if self.stored_group:
            return self.stored_group.get_license_key()
        if self.parent:
            return self.parent.get_license_key()
        if self.copyrights:
            return UNKNOWN_COPYRIGHTED
        return UNKNOWN

    def get_copyright_key(self):
        ''' Identifier corresponding to the copyright holders found '''
        return self.copyrights.key()

    def get_splitting_path(self, options):
        '''Identifier corresponding to the path used for splitting groups.'''
        if self.parent:
            return self.parent.get_splitting_path(options)

    def get_group_key(self, options):
        '''Identifier to split files into groups in the copyright file.'''
        license_key = self.get_license_key()
        if options.group_by == 'copyright':
            copyright_key = self.get_copyright_key()
        else:
            copyright_key = None

        path_key = self.get_splitting_path(options)

        group_key = (license_key, copyright_key, path_key)
        return group_key

    def get_first_untallied(self):
        '''Returns the topmost entry that has not yet been tallied.
           Should not be called with an entry that has already been tallied.
        '''
        assert not self.tallied
        if self.parent and not self.parent.tallied:
            return self.parent.get_first_untallied()
        return self

    def tally(self):
        '''Marks this item and all its parents as tallied.'''
        if self.tallied:
            return
        self.tallied = True
        if self.parent:
            self.parent.tally()

    @property
    def included(self):
        if self._included is not None:
            return self._included
        if self.parent:
            self._included = self.parent.included
            return self._included
        return False

    @included.setter
    def included(self, value):
        self._included = value

    def process_licenses(self, root, options):
        '''Parse the license and copyright information for this file.'''
        # First check if we actually want to process this file
        if options.exclude_fullname_re.match(self.fullname) or \
           options.exclude_special_re.match(self.fullname):
            logging.debug('Ignored file %s (by --exclude)', self.name)
            return
        # Now read it and parse any license or copyright
        fullpath = os.path.join(root, self.fullname)
        copyrights, licenses = parse_file(fullpath, options)
        logging.debug("Adding copyrights %s, to %s", copyrights, self)
        self.add_copyrights(copyrights)
        logging.debug("Resulted in %s", self.copyrights)
        logging.debug("Adding licenses %s, to %s", licenses, self)
        self.add_licenses(licenses)
        logging.debug("Resulted in %s", self.licenses)
        return self


class DirInfo(FileInfo):

    def __init__(self, parent=None, name=''):
        super().__init__(parent, name)
        self.file_list = {}
        # Total number of files in sub tree
        self.total = 0
        self.tasks = set()

    @property
    def parsed_license(self):
        if self.licenses:
            return 'Inherited({})'.format(
                ' or '.join(sorted(self.licenses)))
        return ''

    def add(self, path=None, dirs=None, files=None):

        # Update the total number of files
        self.total += len(dirs) + len(files)

        if path:
            name = path[0]
            if name not in self.file_list:
                self.file_list[name] = DirInfo(self, name)
            return self.file_list[name].add(path[1:], dirs, files)

        for name in dirs:
            if name not in self.file_list:
                self.file_list[name] = DirInfo(self, name)
        for name in files:
            if name not in self.file_list:
                self.file_list[name] = FileInfo(self, name)

    def get_splitting_path(self, options):
        '''Identifier corresponding to the path used for splitting groups.
        When --split-on-license is true, this is the last dirinfo that contains
        a license, when --no-split-on-license is used, it's the root dirinfo.
        Additionally, the debian directory is treated specially when
        --split-debian is true.'''
        if (options.split_on_license and self.licenses) or not self.parent:
            return self.fullname
        if options.split_debian and self.fullname == 'debian':
            return self.fullname
        return self.parent.get_splitting_path(options)

    def __iter__(self):
        yield self
        for _, item in self.file_list.items():
            if isinstance(item, DirInfo):
                for subitem in item:
                    yield subitem
            else:
                yield item

    def process_licenses(self, root, options):
        '''Scan the directory for license files.'''

        dir_licenses = set()
        filenames = set()

        for task in futures.as_completed(self.tasks):
            fileinfo = task.result()
            if not fileinfo:
                continue
            licenses = fileinfo.licenses
            if not licenses:
                # Not a license?
                continue

            dir_licenses.update(licenses)
            filenames.add(fileinfo.fullname)

        self.tasks = set()

        if dir_licenses:
            self.add_licenses(dir_licenses, filenames)

        return self


class RootInfo(DirInfo):

    def __init__(self, root):
        super().__init__()
        self.root = root

    @property
    def names(self):
        for item in self:
            yield str(item)

    def __getitem__(self, key):
        if not key or key == '.':
            return self

        fullpath = key.split(os.path.sep)
        current = self

        for path in fullpath:
            if not isinstance(current, DirInfo) or \
               path not in current.file_list:
                raise IndexError('{} not found'.format(key))
            current = current.file_list[path]
        return current

    @staticmethod
    def build(options):

        tree = RootInfo(root=options.root)

        for root, dirs, files in tqdm(os.walk(options.root, topdown=True),
                                      desc='Building tree',
                                      unit='dir',
                                      dynamic_ncols=True,
                                      disable=(options.debug or
                                               not options.progress)):
            logging.debug('Listing %s', root)
            to_remove = []
            for directory in dirs:
                if options.exclude_directory_re.search(directory):
                    logging.debug('Ignoring directory %s as requested', directory)
                    to_remove.append(directory)
            for d in to_remove:
                dirs.remove(d)
            to_remove = []
            for filename in files:
                if options.exclude_file_re.search(filename):
                    logging.debug('Ignoring file %s as requested', filename)
                    to_remove.append(filename)
                elif os.path.islink(os.path.join(root, filename)):
                    logging.debug('Ignoring symlink %s', filename)
                    # Skip symlinks
                    to_remove.append(filename)

                # elif os.path.getsize(os.path.join(root, filename)) == 0:
                #     # Skip empty files
                #     logging.debug('Skipping empty file %s', filename)
                #     to_remove.append(filename)
            for f in to_remove:
                files.remove(f)

            # Remove the "./"
            base = root[len(tree.root) + 1:]
            tree.add(base.split(os.path.sep) if base else [], dirs, files)

        tree.tag_included_files(options)
        return tree

    def tag_included_files(self, options):
        '''Determine which files to include, from the specified cmd options.'''

        if not options.files:
            self.included = True

        for filename in options.files:
            # Special case, if the root directory was received as a positional
            # parameter, include the whole tree.
            if filename == ".":
                self.included = True
                return

            # Filenames not in the tree will cause an exception. Catch that and
            # log an error in that case.
            try:
                self[filename].included = True
            except IndexError:
                logging.error('%s not in source tree', filename)

    def process(self, options):
        '''Iterate over the full tree processing their licenses.'''
        kwargs = {}
        if options.jobs:
            kwargs['max_workers'] = options.jobs

        known_license_filenames = (
            r'^COPYING',
            r'^COPYRIGHT$',
            r'^LICENSE',
        )

        with futures.ThreadPoolExecutor(**kwargs) as executor:
            tasks = set()
            dirs_with_licenses = set()

            for item in tqdm(self,
                             desc='Queueing',
                             total=self.total,
                             unit='file',
                             dynamic_ncols=True,
                             disable=(options.debug or not options.progress)):

                if isinstance(item, DirInfo):
                    continue

                if not item.included:
                    continue

                task = executor.submit(item.process_licenses, self.root, options)

                for filename_re in known_license_filenames:
                    if not re.match(filename_re, item.name):
                        continue
                    item.parent.tasks.add(task)
                    dirs_with_licenses.add(item.parent)
                    break
                else:
                    tasks.add(task)

            for dirinfo in dirs_with_licenses:
                task = executor.submit(dirinfo.process_licenses, self.root,
                                       options)
                tasks.add(task)

            done_iter = futures.as_completed(tasks)
            for task in tqdm(done_iter,
                             desc='Processing',
                             total=len(tasks),
                             unit='file',
                             dynamic_ncols=True,
                             disable=(options.debug or not options.progress)):
                pass


class FileGroup(object):

    def __init__(self):
        self.files = {}

    def __len__(self):
        return len(self.files)

    def __iter__(self):
        return iter(self.files.values())

    def __repr__(self):
        return 'FileGroup({})'.format(str(self.sorted_members()))

    def add_file(self, fileinfo):
        self.files[fileinfo.fullname] = fileinfo

    def del_file(self, fileinfo):
        return self.files.pop(fileinfo.fullname)

    def get_patterns(self):
        '''Returns the wildcarded patters for the contained files.

        The order in which this method is called is important.  It should be
        called with the group that will get the more general wildcard first, and
        with the groups that will get more specific values later.
        '''
        patterns = set()
        to_tally = set()

        for fileinfo in self:
            untallied = fileinfo.get_first_untallied()
            to_tally.add(untallied)

        for fileinfo in to_tally:
            if isinstance(fileinfo, DirInfo):
                if fileinfo.fullname:
                    patterns.add(fileinfo.fullname + '/*')
                else:
                    patterns.add('*')
            else:
                patterns.add(fileinfo.fullname)
        for fileinfo in self:
            fileinfo.tally()

        return sorted(patterns)

    def key(self, options):
        if not self.files:
            return ''
        return os.path.commonpath(
            tuple(v.get_splitting_path(options) for v in self.files.values()))

    def commonpath(self):
        if not self.files:
            return ''
        return os.path.commonpath(tuple(f.fullname for f in self.files.values()))

    def sorted_members(self):
        return sorted(self.files.keys())


class CopyrightGroup(object):

    def __init__(self):
        self.members = {}
        self.by_email = {}

    def __len__(self):
        return len(self.members)

    def __iter__(self):
        return iter(self.members.values())

    def __repr__(self):
        return 'CopyrightGroup({})'.format(str(list(self.sorted_members())))

    def add(self, copyright_):
        if copyright_.email and copyright_.email in self.by_email:
            previous = self.by_email[copyright_.email]
            return previous.merge(copyright_)
        elif copyright_.person in self.members:
            previous = self.members[copyright_.person]
            return previous.merge(copyright_)

        if copyright_.email:
            self.by_email[copyright_.email] = copyright_
        self.members[copyright_.person] = copyright_
        return copyright_

    def extend(self, copyrights):
        for copyright_ in copyrights:
            self.add(copyright_)

    def key(self):
        return tuple(sorted(
            copyright.email if copyright.email else copyright.person
            for copyright in self.members.values()))

    def merge(self, other):
        for _, value in other.members.items():
            self.add(value)
        return self

    def sorted_members(self):
        return (str(self.members[key]) for key in sorted(self.members.keys()))



#!/usr/bin/env python3
# encoding: utf-8

# Unit tests for decopy
# Copyright (C) 2016 Maximiliano Curia <maxy@debian.org>
#
# License: ISC
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
#  AND FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
#  LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
#  OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
#  PERFORMANCE OF THIS SOFTWARE.

import io
import sys
import unittest

from unittest.mock import DEFAULT, patch

from decopy import cmdoptions
from decopy import output
from decopy import tree


class TestOutput(unittest.TestCase):

    def setUp(self):
        # Create one root with one dir with one file
        self.tree = tree.RootInfo(root='.')
        self.groups = {}

    def testEmpty(self):
        groups = {}
        copyright_ = None
        options = cmdoptions.process_options([])

        with patch.multiple(sys, stdout=DEFAULT, stderr=DEFAULT,
                            new_callable=io.StringIO) as _sys:
            output.generate_output(groups, self.tree, copyright_, options)
        out = _sys['stdout'].getvalue()
        err = _sys['stderr'].getvalue()
        self.assertFalse(len(err), 'Unexpected stderr: {}'.format(err))
        self.assertEqual(out, 'Format: https://www.debian.org/doc/'
                         'packaging-manuals/copyright-format/1.0/\n'
                         'Source: TODO\n',
                         'Unexpected stdout: {}'.format(out))

    def testEmptyPartial(self):
        groups = {}
        copyright_ = None
        options = cmdoptions.process_options(['--mode=partial', '.'])

        with patch.multiple(sys, stdout=DEFAULT, stderr=DEFAULT,
                            new_callable=io.StringIO) as _sys:
            output.generate_output(groups, self.tree, copyright_, options)
        out = _sys['stdout'].getvalue()
        err = _sys['stderr'].getvalue()
        self.assertFalse(len(err), 'Unexpected stderr: {}'.format(err))
        self.assertEqual(out, 'Format: https://www.debian.org/doc/'
                         'packaging-manuals/copyright-format/1.0/\n'
                         'Source: TODO\n'
                         'Comment: *** only: . ***\n',
                         'Unexpected stdout: {}'.format(out))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*- vim60:fdm=marker
#
# Copyright: 2016, Maximiliano Curia <maxy@debian.org>
#
# License: ISC
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
#  AND FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
#  LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
#  OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
#  PERFORMANCE OF THIS SOFTWARE.

'''File Parsers'''

import codecs
import collections
import functools
import html
import io
import logging
import shutil
import string
import subprocess
import tokenize

try:
    import regex as re
except ImportError:
    import re
import re as reo

import xdg.Mime

# Local modules
from .matchers import find_licenses, clean_comments, parse_holders

# initialize xdg cache before using it
xdg.Mime.update_cache()

PRINTABLE_BYTES = bytes(string.printable, encoding='ascii')


def generic_input(reader):

    def detect_bom(buffer):
        # This is what chardet.detect does, but chardet fallsback to ascii. :/
        # So you either need to feed all the file to chardet and read it
        # twice (or keep it in memory)
        for encoding, boms in (
                ('UTF-8-SIG', (codecs.BOM_UTF8,)),
                ('UTF-32', (codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE)),
                ('X-ISO-10646-UCS-4-3412', (b'\xFE\xFF\x00\x00',)),
                ('X-ISO-10646-UCS-4-2143', (b'\x00\x00\xFF\xFE',)),
                ('UTF-16', (codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE)),
        ):
            if buffer.startswith(boms):
                return encoding
        return 'utf-8'

    buffer = reader.peek(4)
    default_encoding = detect_bom(buffer)
    fallback_encoding = 'latin1'

    for raw_line in reader:
        try:
            line = raw_line.decode(default_encoding)
        except UnicodeDecodeError:
            if has_control_chars(raw_line):
                # binary?
                chars = (chr(char) if char in PRINTABLE_BYTES else '\n'
                         for char in raw_line)
                line = ''.join(chars)
                # Drop extra new lines
                line = reo.sub(r'\n{2,}', '\n', line)
            else:
                line = raw_line.decode(fallback_encoding)
        # convert mac to unix
        line = reo.sub(r'\r([^\n])', r'\n\1', line)
        yield line


def _get_content(filename):
    'Get the whole file content'

    content = []

    with open(filename, 'rb') as f:
        for line in generic_input(f):
            content.append(line)

    return ''.join(content)


def _process_content(content):
    'Process file contents'

    holders = parse_holders(content)
    content = clean_comments(content)

    licenses = find_licenses(content)

    return holders, licenses


def generic_parser(filename):
    '''Generic read, treat all files as text

    Binary content is filtered ala 'strings(1)'.
    Bom is detected, files with no bom are treated as utf-8 files, but lines
    that fail to be decoded are decoded with latin1.
    '''

    content = _get_content(filename)
    return _process_content(content)


@functools.lru_cache()
def cmd_parser_factory(*cmd):

    def _parser(filename):

        fullcmd = cmd + (filename,)

        try:
            result = subprocess.run(
                fullcmd, check=True,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            # A bit overkill?
            content = ''.join(
                generic_input(io.BufferedReader(io.BytesIO(result.stdout)))
            )
            content += ''.join(
                generic_input(io.BufferedReader(io.BytesIO(result.stderr)))
            )
        except (subprocess.CalledProcessError, UnicodeDecodeError) as e:
            logging.info('failed to parse %s with %s, (%s)',
                         filename, cmd, e)
            logging.info('falling back')
            return generic_parser(filename)

        return _process_content(content)

    if shutil.which(cmd[0]) is None:
        logging.warn('command %s not found, using generic parser as fallback',
                     cmd[0])
        return generic_parser

    return _parser


def html_parser(filename):
    'Drop/convert html entities, and brs'

    # Keep in mind that most template systems will also match this type. We
    # could use beautifulsoup to obtain the text and the html comments, but we
    # would loose the templating system comments, and the comments of the js
    # parts.

    content = _get_content(filename)
    # Not sure if this should go after the unescape
    content = reo.sub(r'<\s*br\s*/?>', '\n', content)
    try:
        content = html.unescape(content)
    except Exception as e:
        logging.info('failed to unescape html entities from %s, (%s)',
                     filename, e)
        logging.info('using content as is')

    return _process_content(content)


POUND_LINE_RE = re.compile(r'^\s*#')


def pound_lines_parser(filename):
    '''Keep only the lines starting with #'''

    lines = []
    with open(filename, 'rb') as f:

        for line in generic_input(f):
            if POUND_LINE_RE.match(line):
                lines.append(line)
            else:
                lines.append('\n')

    content = ''.join(lines)

    return _process_content(content)


PO_COMMENTED_TRANSLATION_RE = re.compile(r'^\s*#.{,2} msg(?:str|id|ctxt)\b')


def po_file_parser(filename):
    '''Keep only the comments (ignore also #~ msgid lines)'''

    lines = []
    with open(filename, 'rb') as f:

        for line in generic_input(f):
            if not POUND_LINE_RE.match(line):
                lines.append('\n')
            elif PO_COMMENTED_TRANSLATION_RE.match(line):
                lines.append('\n')
            else:
                lines.append(line)

    content = ''.join(lines)

    return _process_content(content)


def python_parser(filename):
    '''Extract comments and doc strings from a python file'''

    lines = []
    newline = True
    with open(filename, 'rb') as f:
        for token in tokenize.tokenize(f.readline):
            if (token.type == tokenize.COMMENT) or \
               (newline and token.type == tokenize.STRING):
                lines.append(token.string)
            elif token.type == tokenize.NEWLINE:
                newline = True
            elif newline and \
               (token.type in {tokenize.INDENT, tokenize.NL, tokenize.ENCODING}):
                continue
            else:
                newline = False
    content = '\n'.join(lines)

    return _process_content(content)


KNOWN_PARSERS = {
    'application/gzip': cmd_parser_factory('zcat'),
    'application/x-bzip': cmd_parser_factory('bzcat'),
    'application/x-lzma': cmd_parser_factory('xzcat'),
    'application/x-xz': cmd_parser_factory('xzcat'),
    'image/svg+xml-compressed': cmd_parser_factory('zcat'),

    # 'application/bzip2': cmd_parser_factory('exiftool'),
    # 'application/json': cmd_parser_factory('exiftool'),
    # 'application/octet-stream': cmd_parser_factory('exiftool'),
    # 'application/postscript': cmd_parser_factory('exiftool'),
    # 'application/rdf+xml': cmd_parser_factory('exiftool'),
    # 'application/x-bittorrent': cmd_parser_factory('exiftool'),
    # 'application/x-gzip': cmd_parser_factory('exiftool'),
    # 'application/x-httpd-php': cmd_parser_factory('exiftool'),
    # 'application/x-rar-compressed': cmd_parser_factory('exiftool'),
    # 'application/x-tar': cmd_parser_factory('exiftool'),
    # 'application/xml': cmd_parser_factory('exiftool'),
    # 'application/zip': cmd_parser_factory('exiftool'),
    # 'image/svg+xml': cmd_parser_factory('exiftool'),
    # 'text/calendar': cmd_parser_factory('exiftool'),
    # 'text/html': cmd_parser_factory('exiftool'),
    # 'text/rtf': cmd_parser_factory('exiftool'),
    # 'text/vcard': cmd_parser_factory('exiftool'),
    'application/ResEdit': cmd_parser_factory('exiftool'),
    'application/dicom': cmd_parser_factory('exiftool'),
    'application/itunes': cmd_parser_factory('exiftool'),
    'application/msword': cmd_parser_factory('exiftool'),
    'application/mxf': cmd_parser_factory('exiftool'),
    'application/pdf': cmd_parser_factory('exiftool'),
    'application/sketch': cmd_parser_factory('exiftool'),
    'application/vnd.adobe.fla': cmd_parser_factory('exiftool'),
    'application/vnd.adobe.illustrator': cmd_parser_factory('exiftool'),
    'application/vnd.adobe.indesign-idml-package': cmd_parser_factory('exiftool'),
    'application/vnd.adobe.photoshop': cmd_parser_factory('exiftool'),
    'application/vnd.iccprofile': cmd_parser_factory('exiftool'),
    'application/vnd.ms-excel': cmd_parser_factory('exiftool'),
    'application/vnd.ms-excel.addin.macroEnabled': cmd_parser_factory('exiftool'),
    'application/vnd.ms-excel.sheet.binary.macroEnabled': cmd_parser_factory('exiftool'),
    'application/vnd.ms-excel.sheet.macroEnabled': cmd_parser_factory('exiftool'),
    'application/vnd.ms-excel.template.macroEnabled': cmd_parser_factory('exiftool'),
    'application/vnd.ms-officetheme': cmd_parser_factory('exiftool'),
    'application/vnd.ms-powerpoint': cmd_parser_factory('exiftool'),
    'application/vnd.ms-powerpoint.presentation.macroEnabled': cmd_parser_factory('exiftool'),
    'application/vnd.ms-powerpoint.slideshow.macroEnabled': cmd_parser_factory('exiftool'),
    'application/vnd.ms-powerpoint.template.macroEnabled': cmd_parser_factory('exiftool'),
    'application/vnd.ms-word.document.macroEnabled': cmd_parser_factory('exiftool'),
    'application/vnd.ms-word.template.macroEnabledTemplate': cmd_parser_factory('exiftool'),
    'application/vnd.oasis.opendocument.chart': cmd_parser_factory('exiftool'),
    'application/vnd.oasis.opendocument.database': cmd_parser_factory('exiftool'),
    'application/vnd.oasis.opendocument.formula': cmd_parser_factory('exiftool'),
    'application/vnd.oasis.opendocument.graphics': cmd_parser_factory('exiftool'),
    'application/vnd.oasis.opendocument.image': cmd_parser_factory('exiftool'),
    'application/vnd.oasis.opendocument.presentation': cmd_parser_factory('exiftool'),
    'application/vnd.oasis.opendocument.spreadsheet': cmd_parser_factory('exiftool'),
    'application/vnd.oasis.opendocument.text': cmd_parser_factory('exiftool'),
    'application/vnd.openxmlformats-officedocument.presentationml.presentation': cmd_parser_factory('exiftool'),
    'application/vnd.openxmlformats-officedocument.presentationml.slideshow': cmd_parser_factory('exiftool'),
    'application/vnd.openxmlformats-officedocument.presentationml.template': cmd_parser_factory('exiftool'),
    'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet': cmd_parser_factory('exiftool'),
    'application/vnd.openxmlformats-officedocument.spreadsheetml.template': cmd_parser_factory('exiftool'),
    'application/vnd.openxmlformats-officedocument.wordprocessingml.document': cmd_parser_factory('exiftool'),
    'application/vnd.openxmlformats-officedocument.wordprocessingml.template': cmd_parser_factory('exiftool'),
    'application/vnd.palm': cmd_parser_factory('exiftool'),
    'application/vnd.rn-realmedia': cmd_parser_factory('exiftool'),
    'application/vnd.rn-realmedia-vbr': cmd_parser_factory('exiftool'),
    'application/x-3ds': cmd_parser_factory('exiftool'),
    'application/x-captureone': cmd_parser_factory('exiftool'),
    'application/x-chm': cmd_parser_factory('exiftool'),
    'application/x-dfont': cmd_parser_factory('exiftool'),
    'application/x-font-otf': cmd_parser_factory('exiftool'),
    'application/x-font-ttf': cmd_parser_factory('exiftool'),
    'application/x-font-type1': cmd_parser_factory('exiftool'),
    'application/x-indesign': cmd_parser_factory('exiftool'),
    'application/x-indesign-interchange': cmd_parser_factory('exiftool'),
    'application/x-iso9660-image': cmd_parser_factory('exiftool'),
    'application/x-iwork-keynote-sffkey': cmd_parser_factory('exiftool'),
    'application/x-iwork-numbers-sffnumbers': cmd_parser_factory('exiftool'),
    'application/x-iwork-pages-sffpages': cmd_parser_factory('exiftool'),
    'application/x-magick-image': cmd_parser_factory('exiftool'),
    'application/x-matroska': cmd_parser_factory('exiftool'),
    'application/x-mie': cmd_parser_factory('exiftool'),
    'application/x-mobipocket-ebook': cmd_parser_factory('exiftool'),
    'application/x-shockwave-flash': cmd_parser_factory('exiftool'),
    'application/x-visio': cmd_parser_factory('exiftool'),
    'application/x-wmf': cmd_parser_factory('exiftool'),
    'audio/audible': cmd_parser_factory('exiftool'),
    'audio/flac': cmd_parser_factory('exiftool'),
    'audio/mpeg': cmd_parser_factory('exiftool'),
    'audio/ogg': cmd_parser_factory('exiftool'),
    'audio/x-aiff': cmd_parser_factory('exiftool'),
    'audio/x-ds2': cmd_parser_factory('exiftool'),
    'audio/x-dss': cmd_parser_factory('exiftool'),
    'audio/x-matroska': cmd_parser_factory('exiftool'),
    'audio/x-monkeys-audio': cmd_parser_factory('exiftool'),
    'audio/x-ms-wma': cmd_parser_factory('exiftool'),
    'audio/x-musepack': cmd_parser_factory('exiftool'),
    'audio/x-pn-realaudio': cmd_parser_factory('exiftool'),
    'audio/x-pn-realaudio-plugin': cmd_parser_factory('exiftool'),
    'font/ttf': cmd_parser_factory('exiftool'),
    'image/apng': cmd_parser_factory('exiftool'),
    'image/bmp': cmd_parser_factory('exiftool'),
    'image/bpg': cmd_parser_factory('exiftool'),
    'image/flif': cmd_parser_factory('exiftool'),
    'image/gif': cmd_parser_factory('exiftool'),
    'image/jng': cmd_parser_factory('exiftool'),
    'image/jp2': cmd_parser_factory('exiftool'),
    'image/jpeg': cmd_parser_factory('exiftool'),
    'image/jpm': cmd_parser_factory('exiftool'),
    'image/jpx': cmd_parser_factory('exiftool'),
    'image/pgf': cmd_parser_factory('exiftool'),
    'image/pict': cmd_parser_factory('exiftool'),
    'image/png': cmd_parser_factory('exiftool'),
    'image/tiff': cmd_parser_factory('exiftool'),
    'image/vnd.djvu': cmd_parser_factory('exiftool'),
    'image/vnd.fpx': cmd_parser_factory('exiftool'),
    'image/vnd.ms-photo': cmd_parser_factory('exiftool'),
    'image/vnd.radiance': cmd_parser_factory('exiftool'),
    'image/x-adobe-dng': cmd_parser_factory('exiftool'),
    'image/x-canon-cr2': cmd_parser_factory('exiftool'),
    'image/x-canon-cr3': cmd_parser_factory('exiftool'),
    'image/x-canon-crw': cmd_parser_factory('exiftool'),
    'image/x-dpx': cmd_parser_factory('exiftool'),
    'image/x-epson-erf': cmd_parser_factory('exiftool'),
    'image/x-exr': cmd_parser_factory('exiftool'),
    'image/x-exv': cmd_parser_factory('exiftool'),
    'image/x-fujifilm-raf': cmd_parser_factory('exiftool'),
    'image/x-gopro-gpr': cmd_parser_factory('exiftool'),
    'image/x-hasselblad-3fr': cmd_parser_factory('exiftool'),
    'image/x-hasselblad-fff': cmd_parser_factory('exiftool'),
    'image/x-j2c': cmd_parser_factory('exiftool'),
    'image/x-kodak-dcr': cmd_parser_factory('exiftool'),
    'image/x-kodak-k25': cmd_parser_factory('exiftool'),
    'image/x-kodak-kdc': cmd_parser_factory('exiftool'),
    'image/x-leica-rwl': cmd_parser_factory('exiftool'),
    'image/x-light-lri': cmd_parser_factory('exiftool'),
    'image/x-lytro-lfp': cmd_parser_factory('exiftool'),
    'image/x-mamiya-mef': cmd_parser_factory('exiftool'),
    'image/x-minolta-mrw': cmd_parser_factory('exiftool'),
    'image/x-nikon-nef': cmd_parser_factory('exiftool'),
    'image/x-nikon-nrw': cmd_parser_factory('exiftool'),
    'image/x-olympus-orf': cmd_parser_factory('exiftool'),
    'image/x-paintshoppro': cmd_parser_factory('exiftool'),
    'image/x-panasonic-rw2': cmd_parser_factory('exiftool'),
    'image/x-pentax-pef': cmd_parser_factory('exiftool'),
    'image/x-photo-cd': cmd_parser_factory('exiftool'),
    'image/x-portable-bitmap': cmd_parser_factory('exiftool'),
    'image/x-portable-graymap': cmd_parser_factory('exiftool'),
    'image/x-portable-pixmap': cmd_parser_factory('exiftool'),
    'image/x-quicktime': cmd_parser_factory('exiftool'),
    'image/x-raw': cmd_parser_factory('exiftool'),
    'image/x-rawzor': cmd_parser_factory('exiftool'),
    'image/x-samsung-srw': cmd_parser_factory('exiftool'),
    'image/x-sigma-x3f': cmd_parser_factory('exiftool'),
    'image/x-sony-arw': cmd_parser_factory('exiftool'),
    'image/x-sony-sr2': cmd_parser_factory('exiftool'),
    'image/x-sony-srf': cmd_parser_factory('exiftool'),
    'image/x-tiff-big': cmd_parser_factory('exiftool'),
    'image/x-xcf': cmd_parser_factory('exiftool'),
    'video/3gp': cmd_parser_factory('exiftool'),
    'video/3gpp': cmd_parser_factory('exiftool'),
    'video/3gpp2': cmd_parser_factory('exiftool'),
    'video/divx': cmd_parser_factory('exiftool'),
    'video/m2ts': cmd_parser_factory('exiftool'),
    'video/mng': cmd_parser_factory('exiftool'),
    'video/mp4': cmd_parser_factory('exiftool'),
    'video/mpeg': cmd_parser_factory('exiftool'),
    'video/ogg': cmd_parser_factory('exiftool'),
    'video/quicktime': cmd_parser_factory('exiftool'),
    'video/vnd.rn-realvideo': cmd_parser_factory('exiftool'),
    'video/webm': cmd_parser_factory('exiftool'),
    'video/x-canon-crm': cmd_parser_factory('exiftool'),
    'video/x-dv': cmd_parser_factory('exiftool'),
    'video/x-flv': cmd_parser_factory('exiftool'),
    'video/x-matroska': cmd_parser_factory('exiftool'),
    'video/x-ms-asf': cmd_parser_factory('exiftool'),
    'video/x-ms-dvr': cmd_parser_factory('exiftool'),
    'video/x-ms-wmv': cmd_parser_factory('exiftool'),
    'video/x-ms-wtv': cmd_parser_factory('exiftool'),
    'video/x-red-r3d': cmd_parser_factory('exiftool'),

    'text/x-gettext-translation': po_file_parser,
    'text/x-gettext-translation-template': po_file_parser,
    'text/x-po': po_file_parser,
    'text/x-pot': po_file_parser,

    'text/x-python': python_parser,

    'text/html': html_parser,
    'application/x-php': html_parser,
}


def has_control_chars(bytes_seq):
    for i in bytes_seq:
        # ascii control chars (C0), and delete (DEL)
        if i < 32 or i == 127:
            # except tab (HT), line feed (LF) and carriage return (CR)
            if i in (9, 10, 13):
                continue
            return True
    return False


def lookup_mimetypes(xdg_type):

    queue = collections.deque()
    queue.append(xdg_type)
    seen = set()
    while queue:
        xdg_mime = queue.popleft()
        mime = str(xdg_mime)

        if mime in seen:
            continue
        seen.add(mime)

        parser = KNOWN_PARSERS.get(mime)
        if parser:
            return parser

        queue.extend(xdg_mime.inherits_from())
        logging.debug("Original type: %s, queue: %s", xdg_type, queue)

    return generic_parser


def parse_file(fullname, options):
    '''Parses the received file with the matching parser.

    Returns:
        The list of copyrights and licenses obtained by the parser.
    '''
    if options.text:
        parser = generic_parser
    else:
        xdg_type = xdg.Mime.get_type(fullname)
        logging.debug('Type for %s is: %s', fullname, xdg_type)

        parser = lookup_mimetypes(xdg_type)

    copyrights, licenses = parser(fullname)
    logging.debug('Parsed %s: %s, %s', fullname, copyrights, licenses)
    return copyrights, licenses



#!/usr/bin/env python3

from setuptools import setup, find_packages

setup(
    name='Decopy',
    version='0.2.4.9',
    author='Maximiliano Curia',
    author_email='maxy@debian.org',
    description='Automatic debian/copyright Generator',
    long_description='''
    Decopy automates writing and updating the debian/copyright file.

    It reads all files in the source tree, analyzes the licenses and copyright
    messages included and generates the corresponding debian/copyright file.
    When the file already exists, decopy parses it to generate a more complete
    output.
    ''',
    url='https://salsa.debian.org/debian/decopy',
    license='ISC',
    requires=['debian', 'tqdm', 'xdg'],
    classifiers=[
        'Environment :: Console',
        'Programming Language :: Python :: 3',
        'Operating System :: POSIX :: Linux',
        'License :: OSI Approved :: ISC License (ISCL)',
    ],
    packages=find_packages(exclude=['tests', 'tests.*']),
    entry_points={
        'console_scripts': ['decopy = decopy.decopy:main'],
    },
    test_suite='tests',
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*- vim60:fdm=marker
#
# Copyright: 2016, Maximiliano Curia <maxy@debian.org>
#
# License: ISC
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
#  AND FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
#  LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
#  OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
#  PERFORMANCE OF THIS SOFTWARE.

''' Generic functions used by the parsers. '''

try:
    import regex as re
except ImportError:
    import re


# Local modules
from .datatypes import CopyrightHolder, YearRange
from .res import (COMMENTS_SUBS,
                  COPYRIGHT_PRE_IGNORE_RES, COPYRIGHT_POST_IGNORE_RES,
                  COPYRIGHT_PRE_INDICATOR_RE, COPYRIGHT_INDICATOR_RE,
                  CRUFT_SUBS,
                  EMAIL_SUBS,
                  NAME_CRUFT_SUBS,
                  ARTISTIC_SUBS,
                  BSD_SUBS,
                  CC_BY_SUBS,
                  CECILL_SUBS,
                  CNRI_SUBS,
                  EFL_SUBS,
                  GNU_VERSION_SUBS,
                  GNU_EXCEPTION_SUBS,
                  GFDL_SUBS,
                  LPPL_SUBS,
                  MIT_SUBS,
                  ZPL_SUBS,
                  LICENSES_RES,
                  COMPILED_RES,
                  )


def clean_comments(text):
    # replace it's way cheaper than re.sub
    for comment in (
            # C
            '/*', '*/',
            # Lisp
            ';;',
            # (HT|X)ML
            '<!--', '-->'):
        text = text.replace(comment, ' ')
    for comment_sub in COMMENTS_SUBS:
        text = comment_sub.pattern.sub(comment_sub.repl, text)

    return text


# Copyright holders handling

# We are discarding parts that are shorter than MIN_PART_LEN
MIN_PART_LEN = 3

# Year parsing
YEAR_RE = re.compile(
    r'\s*(?:[\s:([]*)?(?P<lo>\d{2,4})[]:\s]*'
    r'(?:(?:[-~=–—]|to)[\s:[]*(?P<hi>\d{1,4})[]:\s]*)?[,/)]*',
    re.IGNORECASE)


def _get_year(text, ref=0):
    year = int(text)
    if year < 10 and ref:
        # Handle: 1990-2
        fix = 0 if (ref % 10) < year else 1
        year += ((ref // 10) + fix) * 10
    if year < 50:
        year += 2000
    if year < 100:
        year += 1900
    if year < 1000:
        return None
    return year


def _year_post_ignore(text, match):
    'Make sure we got a year'

    start = max(0, match.start('lo') - 1)
    end = min(len(text), match.end('lo') + 1)
    return bool(
        re.match(
            r'[a-zA-Z]{0}|.{0}[a-zA-Z@]'.format(match.group('lo')),
            text[start:end]))


def _split_by_years(text, years=None):
    'Split string by years found'
    years_list = [] if years is None else [years]
    parts = [] if years is None else ['']
    last_end = 0
    for year_match in YEAR_RE.finditer(text):
        if _year_post_ignore(text, year_match):
            continue

        match_dict = year_match.groupdict()

        low = _get_year(match_dict['lo'])
        if low is None:
            continue

        high = low
        if match_dict.get('hi', None):
            high = _get_year(match_dict['hi'], low)

        part_len = year_match.start(0) - last_end

        if years is None or part_len >= MIN_PART_LEN:
            parts.append(text[last_end:year_match.start(0)])
            years = YearRange(low=low, high=high)
            years_list.append(years)
        else:
            years.add(low)
            years.add(high)

        last_end = year_match.end(0)

    parts.append(text[last_end:])
    return years_list, parts


# Email parsing
EMAIL_RE = re.compile(
    r'(?:^|[,;:\s<>/\\([])\s*(?P<email>[^]),;:\s<>/\\([]+?@[^]),;:\s<>@/\\([]+?)\s*(?:[]),;:\s/\\<>]|$)')


def _get_holders_emails(text):

    holders_emails = []

    # De-cruft email
    for sub in EMAIL_SUBS:
        text = sub.pattern.sub(sub.repl, text)

    last_end = 0
    for email_match in EMAIL_RE.finditer(text):
        email = email_match.group('email')
        name = text[last_end:email_match.start()]
        last_end = email_match.end()
        holders_emails.append((name, email))
    if len(text) - last_end >= MIN_PART_LEN:
        holders_emails.append((text[last_end:], ''))
    return holders_emails


def _process_name_email_pairs(name_email_pairs, holders, years):
    years_used = False
    for name, email in name_email_pairs:
        # Fixup name
        if name is None:
            name = ''
        name = name.strip(''',.;*'"@-–—[]{} \t''')
        for sub in NAME_CRUFT_SUBS:
            name = sub.pattern.sub(sub.repl, name)
        if len(name) < MIN_PART_LEN:
            name = ''
        if not name and not email:
            if years and len(holders) and not holders[-1].years:
                holders[-1].years.merge(years)
                years = None
            continue
        if (not name and not years) \
                and (email and len(holders) and not holders[-1].email):
            holders[-1].email = email
            continue
        if (not email and not years) \
                and (name and len(holders) and not holders[-1].name):
            holders[-1].name = name
            continue
        holders.append(CopyrightHolder(name, email, years))
        years_used = True
    return None if years_used else years


def _get_holders_years(text, years=None):
    'Get pairs [(years, text)]'
    years_list, parts = _split_by_years(text, years)
    if len(parts[0]) < MIN_PART_LEN:
        return zip(years_list, parts[1:])
    if len(parts[-1]) < MIN_PART_LEN:
        return zip(years_list, parts)
    years_list = [YearRange()] + years_list
    return zip(years_list, parts)


HOLDER_RE = re.compile(r'\s*(?:by\s*)?(?P<holder>\S.*?\S)[\s"\*,;/]*$', re.I)


# Copyright holders factory
def get_copyright_holders(copyright_, years=None):

    years_text_pairs = _get_holders_years(copyright_, years)

    holders = []
    # To be merged or leftover to be returned
    pending_years = None

    for years, text in years_text_pairs:
        if years and pending_years:
            pending_years.merge(years)
        elif pending_years is None or years:
            pending_years = years
        match = HOLDER_RE.match(text)
        if not match:
            # Discard current text, but keep what we found so far
            if pending_years and len(holders) and not holders[-1].years:
                holders[-1].years.merge(pending_years)
                pending_years = None
            continue

        stripped_text = match.group('holder')
        name_email_pairs = _get_holders_emails(stripped_text)

        pending_years = _process_name_email_pairs(
            name_email_pairs, holders, pending_years)

    return holders, pending_years


def _parse_copyright_continuation(text, continuation, years=None):

    if not continuation:
        return '', None
    prefix = continuation.string[:continuation.start(continuation.lastindex)]
    # Special case for C style multiline comments
    alt_prefix = prefix.replace('/*', ' *')
    # normalize spaces
    alt_prefix = alt_prefix.replace('\t', '  ')
    alt_text = text.replace('\t', '  ')
    if not text.startswith(prefix) and not alt_text.startswith(alt_prefix):
        return '', None
    cont_match = re.match(r'\s+', text[len(prefix):]) \
        or re.match(r'\s+', alt_text[len(alt_prefix):])
    if not cont_match and years:
        rest = text[len(prefix):]
    elif not cont_match:
        return '', None
    else:
        rest = cont_match.string[cont_match.end():]
    # If there is nothing else but empty text, drop the continuation
    if re.match(r'\s*$', rest):
        return '', None
    match = continuation
    return rest, match


def _re_split_iter(pattern, text):
    end = 0
    for indicator_match in COPYRIGHT_INDICATOR_RE.finditer(text):
        part = text[end:indicator_match.start()]
        end = indicator_match.end()
        yield part
    yield text[end:]


def parse_copyright(text, continuation=None, years=None):

    if len(text) < 3:
        return None, None

    match = COPYRIGHT_INDICATOR_RE.search(text)
    if match:
        rest = match.string[match.end():]
    else:
        # Process continuations
        rest, match = _parse_copyright_continuation(text, continuation, years)
        if not match:
            return None, None

    # if COPYRIGHT_PRE_IGNORE_RE.search(text) or
    #         COPYRIGHT_POST_IGNORE_RE.search(rest)
    if any(
        ignore_pattern.search(text)
        for ignore_pattern in COPYRIGHT_PRE_IGNORE_RES
    ) or any(
        ignore_pattern.search(rest)
        for ignore_pattern in COPYRIGHT_POST_IGNORE_RES
    ):
        return None, None

    copyrights = []

    for i, part in enumerate(_re_split_iter(COPYRIGHT_INDICATOR_RE, rest)):
        if not part:
            continue
        for sub in CRUFT_SUBS:
            part = sub.pattern.sub(sub.repl, part)
        if part and len(part) > 2:
            copyrights.append(part)

    return copyrights, match


def _add_new_holders(holders, copyrights, years):

    # copyrights is a list of strings, that were split by copyright
    # indicators.
    for copyright_ in copyrights:
        new_holders, years = get_copyright_holders(copyright_, years)
        if new_holders:
            holders.extend(new_holders)
        elif years:
            # If we found a year, but not a holder, maybe the previous
            # holder is missing the years
            # Example: Copyright Maximiliano Curia (c) 2018
            if len(holders) and not holders[-1].years:
                holders[-1].years.merge(years)
                years = None
    # The years might still be relevant for the next copyright holder.
    return years


def _iter_lines(text, start=0):
    ''' Line iterator '''

    end = start - 1
    while True:
        start = end + 1
        end = text.find('\n', start)
        if end <= 0:
            break
        yield text[start:end], start, end
    if text[start:]:
        yield text[start:], start, len(text)


def parse_holders(content):
    ''' Find the copyright holders specified in content '''

    # list containing the copyrightHolders found
    holders = []

    # last position checked
    last_seen = -1

    # Iterate only for lines that might have a copyright indicator.
    for copyright_match in COPYRIGHT_PRE_INDICATOR_RE.finditer(content):
        # Already seen
        if copyright_match.start() < last_seen:
            continue

        match_start = copyright_match.start()
        line_start = content.rfind('\n', 0, match_start) + 1

        # Initialize the parsing state (continuation and years)
        # continuation is set to the last copyright indicator found, that way we
        # can check the following lines, and correctly detect multiple copyright
        # holders after a single copyright indicator. Example:
        # Copyright:
        #  Maximiliano Curia
        #  Margarita Manterola
        continuation = None
        # Keep years, to parse things like:
        #  Copyright 2018
        #    Maximiliano Curia
        years = None

        for line, _, last_seen in _iter_lines(content, line_start):
            copyrights, continuation = parse_copyright(line, continuation, years)
            # An empty continuation means that the copyright block ended.
            if not continuation:
                # remain calm and carry on in the outside iteration
                break
            if not copyrights:
                # A line without indicators, drop the cached years
                years = None
                continue

            # parse copyrights (list of strings) and create CopyrightHolders
            # with that information. Keep the years for the next iteration.
            years = _add_new_holders(holders, copyrights, years)

    return holders


# License handling

def _parse_subs_first(text, fallback, subs):
    for sub in subs:
        match = sub.pattern.search(text)
        if not match:
            continue
        return match.expand(sub.repl).rstrip('.0')

    return fallback


def _parse_subs_all(text, subs):
    found = []
    for sub in subs:
        match = sub.pattern.search(text)
        if not match:
            continue
        found.append(match.expand(sub.repl).rstrip('.0'))

    return found


def parse_artistic(text, match, group):
    return _parse_subs_first(text, group, ARTISTIC_SUBS)


def parse_bsd(text, match, group):
    return _parse_subs_first(text, group, BSD_SUBS)


def parse_cc_by(text, match, group):
    details = _parse_subs_all(text, CC_BY_SUBS)
    return group + ''.join(details)


def parse_cecill(text, match, group):
    return _parse_subs_first(text, group, CECILL_SUBS)


def parse_cnri(text, match, group):
    return _parse_subs_first(text, group, CNRI_SUBS)


def parse_efl(text, match, group):
    return _parse_subs_first(text, group, EFL_SUBS)


def parse_gpl(text, match, group):
    version = _parse_subs_first(text, '', GNU_VERSION_SUBS)

    exceptions = _parse_subs_all(text, GNU_EXCEPTION_SUBS)

    exception = ' with {} exception'.format(
        '_'.join(sorted(exceptions))) if exceptions else ''

    return '{}{}{}'.format(group,
                           '-{}'.format(version.rstrip('0.')) if version else '',
                           exception)


def parse_gfdl(text, match, group):
    details = _parse_subs_all(text, GFDL_SUBS)
    return group + ''.join(details)


def parse_lppl(text, match, group):
    details = _parse_subs_all(text, LPPL_SUBS)
    return group + ''.join(details)


def parse_mit(text, match, group):
    return _parse_subs_first(text, group, MIT_SUBS)


def parse_zpl(text, match, group):
    details = _parse_subs_all(text, ZPL_SUBS)
    return group + ''.join(details)


PARSE_DETAILS = {
    'Artistic': parse_artistic,
    'BSD': parse_bsd,
    'CC-BY': parse_cc_by,
    'GPL': parse_gpl,
    'LGPL': parse_gpl,
    'GFDL': parse_gfdl,
    'LPPL': parse_lppl,
    'ZPL': parse_zpl,
    'AGPL': parse_gpl,
    'CeCILL': parse_cecill,
    'CNRI': parse_cnri,
    'EFL': parse_efl,
    'MIT/X11': parse_mit,
}


def find_licenses(text):
    '''Scan the text for presence of any of the supported licenses.

    Returns:
        A list of the license ids found in the text.
    '''
    licenses = {}

    for i, license_re in enumerate(LICENSES_RES):
        # Keep the license name in this frame for easier debugging
        license_ = license_re.license
        if license_re.license in licenses:
            continue
        match = COMPILED_RES[i].search(text)
        if not match:
            continue
        if license_re.get_detail:
            license_ = license_re.get_detail(text, match, license_)
        elif license_ in PARSE_DETAILS:
            license_ = PARSE_DETAILS[license_](text, match, license_)
        licenses[license_re.license] = license_

    return licenses.values()



#!/usr/bin/env python3
# -*- coding: utf-8 -*- vim60:fdm=marker
#
# Copyright: 2016, Maximiliano Curia <maxy@debian.org>
#
# License: ISC
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
#  AND FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
#  LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
#  OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
#  PERFORMANCE OF THIS SOFTWARE.

# Local modules
from .tree import RootInfo, DirInfo
from .dep5 import Copyright, Group
from .output import generate_output
from .cmdoptions import process_options


def prepare_output_groups(filetree, copyright_, options):

    copyright_.remove_misplaced_files(options)
    groups = copyright_.get_group_dict(options)

    for fileinfo in filetree:
        if options.mode == 'partial' and not fileinfo.included:
            continue

        # if fileinfo.group is set, it means that the parsed group matches the
        # stored group and there's no need to update it
        if fileinfo.group:
            continue
        if isinstance(fileinfo, DirInfo):
            continue

        # If we reach this it means the file needs to be added to a group.
        # TODO: for diff mode
        file_key = fileinfo.get_group_key(options)
        group = groups.setdefault(file_key, Group(file_key))
        group.add_file(fileinfo)
        fileinfo.group = group

    return groups


def main():

    options = process_options()

    filetree = RootInfo.build(options)
    copyright_ = Copyright.build(filetree, options)

    copyright_.process(filetree)
    filetree.process(options)

    groups = prepare_output_groups(filetree, copyright_, options)

    generate_output(groups, filetree, copyright_, options)

    return 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*- vim60:fdm=marker
#
# Copyright: 2016, Maximiliano Curia <maxy@debian.org>
#
# License: ISC
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#  .
#  THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
#  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
#  AND FITNESS. IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
#  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
#  LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
#  OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
#  PERFORMANCE OF THIS SOFTWARE.

''' Command line option handling. '''

import argparse
import logging
import os

try:
    import regex as re
except ImportError:
    import re


class Defaults(object):

    exclude_file_regex = r'''
    # Ignore general backup files
    ~$|
    # Ignore emacs recovery files
    \.\#.*$|
    # Ignore vi swap files
    \..*\.swp$|
    # Ignore baz-style junk files or directories
    ^,,|
    # File-names that should be ignored (never directories)
    ^(?:DEADJOE|\.cvsignore|\.arch-inventory|\.bzrignore|\.gitignore)$|
    # File or directory names that should be ignored
    ^(?:CVS|RCS|\.deps|\{arch\}|\.arch-ids|\.svn|\.hg|_darcs|\.git|
    \.shelf|_MTN|\.bzr(?:\.backup|tags)?)$
    '''

    exclude_directory_regex = r'''
    # Ignore baz-style junk files or directories
    ^,,|
    # File or directory names that should be ignored
    ^(?:CVS|RCS|\.deps|\{arch\}|\.arch-ids|\.svn|\.hg|_darcs|\.git|
    \.pc|\.shelf|_MTN|\.bzr(?:\.backup|tags)?)$|
    # Python compiled files
    ^__pycache__$
    '''

    # Exclude debian/copyright and debian/changelog by default
    exclude_special_regex = r'''
    ^debian/(copyright|changelog)$
    '''

    # Empty exclude regexp, to be overridden by users
    exclude_fullname_regex = r'''
    ^$
    '''

    # Top directory for processing files
    root = '.'

    # Files/dirs to parse
    files = []

    # Mode to work on:
    # partial: only process the files/dirs specified and generate the output
    #          for this subset
    # full: (default) parse the files/dirs specified, and generate the
    #       copyright information for the whole source tree
    mode = 'full'

    # copyright 1.0 file, relative to the package dir
    copyright_file = 'debian/copyright'

    # group by license or by copyright
    group_by = 'license'

    # Split groups on sub dirs containing licenses
    split_on_license = True

    # Split the debian/* file paragraph
    split_debian = True

    # Treat all files as text
    text = False

    # Try to use glob patterns in the File: output
    glob = True

    # Progress bar
    progress = True

    # Number of parallel jobs (0 for auto)
    jobs = 0

    # output filename, empty for stdout
    output = ''


def _add_boolean_argument(parser, name, default, dest=None):
    '''Helper to add options --name --no-name and set the default'''
    if dest is None:
        dest = name.replace('-', '_')
    parser.add_argument('--{}'.format(name), action='store_true',
                        dest=dest)
    parser.add_argument('--no-{}'.format(name),
                        action='store_false', dest=dest)
    parser.set_defaults(**{dest: default})


def process_options(args=None):

    def _filename(arg):
        if not os.path.exists(arg):
            raise argparse.ArgumentTypeError("{} does not exist".format(arg))
        return arg

    kwargs = {
        'format': '[%(levelname)s] %(message)s',
    }

    parser = argparse.ArgumentParser(
        description='License checker and copyright dep5 helper',
        fromfile_prefix_chars='@'
    )
    parser.add_argument('-X', '--exclude', default=Defaults.exclude_fullname_regex)

    parser.add_argument('--mode', choices=['full', 'partial'],
                        default=Defaults.mode)

    parser.add_argument('--copyright-file',
                        default=Defaults.copyright_file)

    parser.add_argument('--debug', action='store_true')
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('-q', '--quiet', action='store_true')

    parser.add_argument('-a', '--text', action='store_true')

    parser.add_argument('--group-by', choices=['license', 'copyright'],
                        default=Defaults.group_by)
    _add_boolean_argument(parser, 'split-on-license',
                          Defaults.split_on_license)
    _add_boolean_argument(parser, 'split-debian',
                          Defaults.split_debian)

    _add_boolean_argument(parser, 'glob',
                          Defaults.glob)

    _add_boolean_argument(parser, 'progress',
                          Defaults.progress)
    parser.add_argument('--jobs', '-j', type=int, default=Defaults.jobs,
                        help='Number of jobs to run in parallel (0 for auto)')

    parser.add_argument('-o', '--output', default=Defaults.output)

    parser.add_argument('--root', default=Defaults.root,
                        type=_filename)

    parser.add_argument('files', nargs='*', default=Defaults.files,
                        type=_filename)

    options = parser.parse_args(args)

    if options.debug:
        kwargs['level'] = logging.DEBUG
    elif options.verbose:
        kwargs['level'] = logging.INFO
    elif options.quiet:
        kwargs['level'] = logging.ERROR
    else:
        kwargs['level'] = logging.WARN

    logging.basicConfig(**kwargs)

    options.exclude_fullname_re = re.compile(options.exclude, re.IGNORECASE | re.VERBOSE)
    options.exclude_special_re = re.compile(Defaults.exclude_special_regex,
                                            re.IGNORECASE | re.VERBOSE)
    options.exclude_file_re = re.compile(Defaults.exclude_file_regex, re.VERBOSE)
    options.exclude_directory_re = re.compile(Defaults.exclude_directory_regex, re.VERBOSE)

    # - can also be used to specify the stdout
    if options.output == '-':
        options.output = ''

    # Force an absolute path for options.root, and sanitize the user option
    # (see #855778, for problem caused by a trailing /)
    options.root = os.path.abspath(options.root)
    # Transform positional arguments into paths that are relative to the root
    # path
    options.files = [os.path.relpath(i, options.root) for i in options.files]

    return options



# This file is part of decopy test data.  It exists only for testing purposes.

# Copyright: 1993-2004, Clara Oswald <clara@example.com>

# Permission to use, copy, modify, distribute, and sell this software and its
# documentation for any purpose is hereby granted without fee, provided that
# the above copyright notice appear in all copies and that both that
# copyright notice and this permission notice appear in supporting
# documentation.  No representations are made about the suitability of this
# software for any purpose.  It is provided "as is" without express or
# implied warranty.
